#include "GenerateIvy.h"

#include <I3DExVertexPrimitive.h>
#include <I3DExVertex.h>
#include <IPolymesh.h>
#include <IPolygon.h>

extern IChangeManager*	gChangeManager;

const MCGUID CLSID_GenerateIvy(R_CLSID_GenerateIvy);
const MCGUID CLSID_GenerateIvyDlg(R_CLSID_GenerateIvyDlg);



void Extension3DInit(IMCUnknown* utilities)
{
	// Perform your dll initialization here
}


void Extension3DCleanup()
{
	// Perform any nec clean-up here
}


TBasicUnknown* MakeCOMObject(const MCCLSID& classId)	// This method instanciate
{														// the object COMBack
	TBasicUnknown* res = NULL;
	if (classId == CLSID_GenerateIvy) res = new GenerateIvyCommand;
	else if (classId == CLSID_GenerateIvyDlg) res = new GenerateIvyDlg;
	MCAssert(res);
	return res;
}


GenerateIvyCommand::GenerateIvyCommand()
{
	// Protected variables
	fSceneDocument = NULL;
	fScene = NULL;
}


GenerateIvyCommand::~GenerateIvyCommand()
{
}



// TBasicSceneCommand methods :
class IsSelectionNonEmpty : public TBasicMenuCallBack
{
protected:
	IsSelectionNonEmpty() {}
	virtual ~IsSelectionNonEmpty() {}
public:
	void static Create(ISelfPrepareMenuCallBack** out);
	virtual boolean MCCOMAPI SelfPrepareMenu(ISceneDocument* sceneDocument);
};


void IsSelectionNonEmpty::Create(ISelfPrepareMenuCallBack** out)
{
	TMCCountedCreateHelper<ISelfPrepareMenuCallBack> result(out);
	result = new IsSelectionNonEmpty();
}


boolean IsSelectionNonEmpty::SelfPrepareMenu(ISceneDocument* sceneDocument)
{
	if (!sceneDocument) return false;
	TMCCountedPtr<ISceneSelection> selection;
	sceneDocument->GetSceneSelection(&selection);

	if (!selection) return false;

	// check that there is at least one object selected other than the universe
	TTreeSelectionIterator selectionIter(selection);
	I3DShTreeElement* tree= selectionIter.First();
	while (tree)
	{
		TMCCountedPtr<I3DShTreeElement> parent;
		tree->GetTop(&parent);
		if (parent) return true;
		tree = selectionIter.Next(); // here, we make sure that this tree element is in the triple buffer
	}

	return false;
}


void GenerateIvyCommand::GetMenuCallBack(ISelfPrepareMenuCallBack** callBack)
{
	IsSelectionNonEmpty::Create(callBack);
}


MCCOMErr GenerateIvyCommand::Init(ISceneDocument* sceneDocument)
{
	fSceneDocument = sceneDocument;
	sceneDocument -> GetSceneSelection(&fSelection);
	fSelection -> Clone(&fCloneSelection);
	return MC_S_OK;
}


MCCOMErr GenerateIvyCommand::Prepare()
{
	if (fSceneDocument)
	{
		fSceneDocument->GetScene(&fScene);

		fScene->GetTreePropertyChangeChannel(&fTreePropertyChannel);
		ThrowIfNil(fTreePropertyChannel);

		fSceneDocument->GetSceneSelectionChannel(&fSelectionChannel);
		ThrowIfNil(fSelectionChannel);
	}

	if (!fTree)
	{
		TMCCountedPtr<I3DShGroup> group;
		fScene->GetTreeRoot(&group);
		if (group->QueryInterface(IID_I3DShTreeElement, (void **) &fTree) != MC_S_OK)
			return MC_S_OK;
	}
	return MC_S_OK;
}




boolean GenerateIvyCommand::Do()
{
	TTreeSelectionIterator iter(fCloneSelection);

	// Clear the selection
	TTreeSelectionIterator selectionIter(fSelection);
	I3DShTreeElement* tree= selectionIter.First();
	while (tree)
	{
		tree->PostMoveChange();
		tree = selectionIter.Next();
	}
	fSelection->ClearSelection();

	// Create the ivy
	I3DShTreeElement* iTree= iter.First();
	while (iTree)
	{
		doIvy(iTree);
		iTree= iter.Next();
	}

	// Third, tell the Shell what object must be redrawn (it can not guess, but we easily can tell it)
	// Note :	using the selection-change-channel makes the interactive renderer add this I3DShTreeElement to the triple buffer, 
	//			and therefore redraws correctly.
	selectionIter = TTreeSelectionIterator(fSelection);
	tree = selectionIter.First();
	while (tree)
	{
		tree->PostMoveChange(); // here, we make sure that this tree element is in the triple buffer
		tree = selectionIter.Next();
	}

	gChangeManager->PostChange(fSelectionChannel, 0, fSelection);

	return true;
}



boolean GenerateIvyCommand::CanUndo()
{
	return false;
}




boolean GenerateIvyCommand::Undo()
{
	return false;
}




boolean GenerateIvyCommand::Redo()
{
	return false;
}



void GenerateIvyCommand::doIvy(I3DShTreeElement *growOnMe)
{
	TMCCountedPtr<I3DShInstance> instance;
	TMCCountedPtr<I3DShTreeElement> instanceTree;
	TMCCountedPtr<IMCUnknown> progressKey;

	TMCCountedPtr<I3DShPrimitiveInstance> prim;
	growOnMe->QueryInterface(IID_I3DShPrimitiveInstance, (void **)(&prim));
	if(!prim)  
	{
		I3DShTreeElement *temp = gShell3DUtilities->GetTreeToEdit(growOnMe);
		temp->QueryInterface(IID_I3DShPrimitiveInstance, (void **)(&prim));
		if(!prim) return;
	}

	myCommon.ivy.myCommon = &myCommon;
	myCommon.mesh.reset();
	myCommon.ivy.reset();
	myCommon.ivy.resetSettings(fData);

	TMCCountedPtr<FacetMesh> outMesh(prim->GetThisInstance()->GetRenderingFacetMesh());
	if(!outMesh) return;
	
	gShellUtilities->BeginProgress(TMCString255("Extract Object Mesh Vertex List"), &progressKey, outMesh->fVertices.GetElemCount());
	for(uint32 x = 0; x < outMesh->fVertices.GetElemCount(); x++)
	{
		BasicVertex vtx;
		vtx.pos.x = outMesh->fVertices[x].X() * 12; // The FacetMesh is in FEET
		vtx.pos.y = outMesh->fVertices[x].Z() * 12; // The FacetMesh is in FEET
		vtx.pos.z = outMesh->fVertices[x].Y() * 12; // The FacetMesh is in FEET
		myCommon.mesh.vertices.push_back(vtx);
		gShellUtilities->IncrementProgress(1, progressKey);
	}
	gShellUtilities->EndProgress(progressKey);

	gShellUtilities->BeginProgress(TMCString255("Extract Object Mesh Facet List"), &progressKey, outMesh->fFacets.GetElemCount());
	for(uint32 x = 0; x < outMesh->fFacets.GetElemCount(); x++)
	{
		BasicTriangle tri;

		// Was 0, 1, 2
		tri.v2id = outMesh->fFacets[x].pt1 + 1; // add one because the ivy code will subtract one later
		tri.v1id = outMesh->fFacets[x].pt2 + 1;
		tri.v0id = outMesh->fFacets[x].pt3 + 1;

		myCommon.mesh.triangles.push_back(tri);
		gShellUtilities->IncrementProgress(1, progressKey);
	}
	gShellUtilities->EndProgress(progressKey);
	
	myCommon.mesh.prepareData();
	myCommon.mesh.calculateVertexNormals();
	myCommon.mesh.prepareData();

	TTreeTransform srcMat;
	growOnMe->GetLocalTreeTransform(srcMat);
	TVector3 off = srcMat.GetHotPoint();

	Vector3d seedPoint;
	seedPoint.x = off.X() * 12.0f;
	seedPoint.y = off.Z() * 12.0f;
	seedPoint.z = off.Y() * 12.0f;

	myCommon.ivy.seed(seedPoint);

	gShellUtilities->BeginProgress(TMCString255("Calculating Ivy Growth"), &progressKey, fData.growSteps);
	for(int64 x = 0; x < fData.growSteps; x++)
	{
		myCommon.ivy.grow();
		gShellUtilities->IncrementProgress(1, progressKey);
		if (myCommon.ivy.aliveCount == 0) break;
	}
	gShellUtilities->EndProgress(progressKey);

	myCommon.ivy.birth();

#if 0 // OLD!
	// Ok, we have results.  Time to put it in the scene!
	TMCCountedPtr<I3DShObject> shObject;
	TMCCountedPtr<I3DShFacetMeshPrimitive> facetMeshPrimitive;
	gShell3DUtilities->CreatePrimitiveByID(kFacetMeshPrimitiveID, &shObject, IID_I3DShFacetMeshPrimitive, (void**) &facetMeshPrimitive);
	ThrowIfNil(facetMeshPrimitive);

	// fill in the mesh and give it to the primitive
	TMCCountedPtr<FacetMesh> theMesh;
	FacetMesh::Create(&theMesh);
	ThrowIfNil(theMesh);
			
	// Set the number of Points and Facets
	theMesh->SetVerticesCount(myCommon.ivy.vertices.size());
	theMesh->SetFacetsCount(myCommon.ivy.triangles.size());

	gShellUtilities->BeginProgress(TMCString255("Build FacetMesh Vertex List"), &progressKey, myCommon.ivy.vertices.size());
	for(int x = 0; x < myCommon.ivy.vertices.size(); x++)
	{
		TVector3 Point;

		Point.X() = myCommon.ivy.vertices[x].pos.x;
		Point.Y() = myCommon.ivy.vertices[x].pos.z;
		Point.Z() = myCommon.ivy.vertices[x].pos.y;
		(theMesh->fVertices).SetElem(x, Point);

		TVector2 UVPoint;
		UVPoint.X() = myCommon.ivy.texCoords[x].pos.x;
		UVPoint.Y() = myCommon.ivy.texCoords[x].pos.y;
		(theMesh->fuv).SetElem(x, UVPoint);
		gShellUtilities->IncrementProgress(1, progressKey);
	}
	gShellUtilities->EndProgress(progressKey);
	gShellUtilities->BeginProgress(TMCString255("Build FacetMesh Facet List"), &progressKey, myCommon.ivy.triangles.size());
	for(int x = 0; x < myCommon.ivy.triangles.size(); x++)
	{
		Triangle trFacet;
		trFacet.pt1 = myCommon.ivy.triangles[x].v0id - 1;
		trFacet.pt2 = myCommon.ivy.triangles[x].v1id - 1;
		trFacet.pt3 = myCommon.ivy.triangles[x].v2id - 1;
		(theMesh->fFacets).SetElem(x, trFacet);

		(theMesh->fUVSpaceID).SetElem(x, myCommon.ivy.triangles[x].matid - 1);
		gShellUtilities->IncrementProgress(1, progressKey);
	}
	gShellUtilities->EndProgress(progressKey);

	TBBox3D bbox;
	theMesh->BuildEdgeList();
	theMesh->CalcPolygonEqnList();
	theMesh->CalcBBox(bbox);

	facetMeshPrimitive->SetNbrLOD(1); // Must be called before 'SetFacetMesh'
	facetMeshPrimitive->SetUVSpaceCount(3);

	UVSpaceInfo uvInfo;
	uvInfo.fWraparound[0] = 0x00;
	uvInfo.fWraparound[1] = 0x00;

	uvInfo.fID = 0;
	uvInfo.fName.FromCAsciiPtr("leaf_adult");
	facetMeshPrimitive->SetUVSpace(0, uvInfo);

	uvInfo.fID = 1;
	uvInfo.fName.FromCAsciiPtr("leaf_young");
	facetMeshPrimitive->SetUVSpace(1, uvInfo);

	uvInfo.fID = 2;
	uvInfo.fName.FromCAsciiPtr("branch");
	facetMeshPrimitive->SetUVSpace(2, uvInfo);

	facetMeshPrimitive->SetFacetMesh(0, 0, theMesh);

	facetMeshPrimitive->SetBoundingBox(bbox);
#endif
	
	// New - build VertexPrimitive directly
	TMCCountedPtr<I3DShObject> shObject;
	TMCCountedPtr<I3DExVertexPrimitive> vtxPrim;
	gShell3DUtilities->CreatePrimitiveByID('vmda', &shObject, IID_I3DExVertexPrimitive, (void **)(&vtxPrim));

	TMCString63 domain;
	
	int32 shaders[3];
	
	domain.FromCPtr("leaf_adult");
	shaders[0] = vtxPrim->AddShadingDomain(domain, false);
	domain.FromCPtr("leaf_young");
	shaders[1] = vtxPrim->AddShadingDomain(domain, false);
	domain.FromCPtr("branch");
	shaders[2] = vtxPrim->AddShadingDomain(domain, false);

	TMCCountedPtr<IPolymesh> polymesh;
	vtxPrim->CreatePolymesh(&polymesh);

	gShellUtilities->BeginProgress(TMCString255("Build Mesh Vertex List"), &progressKey, fData.growSteps);
	TMCCountedPtr<I3DExVertex> vertex;
	for(int x = 0; x < myCommon.ivy.vertices.size(); x++)
	{
		TVector3 Point;
		Point.X() = myCommon.ivy.vertices[x].pos.x;
		Point.Y() = myCommon.ivy.vertices[x].pos.z;
		Point.Z() = myCommon.ivy.vertices[x].pos.y;

		TVector2 UVPoint;
		UVPoint.X() = myCommon.ivy.texCoords[x].pos.x == 0.0 ? 0.00001 : 0.999999;
		UVPoint.Y() = myCommon.ivy.texCoords[x].pos.y == 0.0 ? 0.00001 : 0.999999;

		TVector3 Normal;
		Normal.X() = myCommon.ivy.normals[x].dir.x;
		Normal.Y() = myCommon.ivy.normals[x].dir.y;
		Normal.Z() = myCommon.ivy.normals[x].dir.z;

		TVertex3D newVtx;
		newVtx.fVertex = Point;
		newVtx.fUV = UVPoint;
		newVtx.fNormal = Normal;

		TVMIVertexPtr newVtxPtr = polymesh->ImportVertex(newVtx, true, x);
		newVtxPtr.fVertex->SetSelection(true);

		gShellUtilities->IncrementProgress(1, progressKey);
	}

	gShellUtilities->EndProgress(progressKey);
	gShellUtilities->BeginProgress(TMCString255("Build Mesh Facet List"), &progressKey, fData.growSteps);
	for(int x = 0; x < myCommon.ivy.triangles.size(); x++)
	{
		TMCClassArray<TVMIVertexPtr> vertices;
		TVMIVertexPtr ptrTemp;

		ptrTemp.fUVIndex = 0;
		ptrTemp.fVertex = polymesh->GetVertexNoAddRef(myCommon.ivy.triangles[x].v0id - 1);		
		vertices.AddElem(ptrTemp);

		ptrTemp.fUVIndex = 0;
		ptrTemp.fVertex = polymesh->GetVertexNoAddRef(myCommon.ivy.triangles[x].v1id - 1);		
		vertices.AddElem(ptrTemp);

		ptrTemp.fUVIndex = 0;
		ptrTemp.fVertex = polymesh->GetVertexNoAddRef(myCommon.ivy.triangles[x].v2id - 1);		
		vertices.AddElem(ptrTemp);

		TMCCountedPtr<IPolygon> polygon(polymesh->MakePolygon(vertices));
		polygon->SetShadingDomainID(shaders[myCommon.ivy.triangles[x].matid - 1]);	
	}
	gShellUtilities->EndProgress(progressKey);

	polymesh->SetUVWraparound(false, false);

	gComponentUtilities->CoCreateInstance(CLSID_StandardInstance, NULL,
		MC_CLSCTX_INPROC_SERVER, IID_I3DShInstance, (void**) &instance);

	shObject->SetName(TMCString15("Ivy"));
	fScene->Insert3DObject(shObject);
	instance->Set3DObject(shObject);
	TMCCountedPtr<I3DShMasterShader> mShader;
	fScene->GetDefaultMasterShader(&mShader);
	instance->SetShader(mShader);

	instance->QueryInterface(IID_I3DShTreeElement, (void**) &instanceTree);
	TMCCountedPtr<I3DShMasterGroup> masterGroup;
	fScene->QueryInterface(IID_I3DShMasterGroup,(void**)&masterGroup);
	MCAssert(masterGroup);
	instanceTree->SetMasterGroup(masterGroup);

	instanceTree->CenterHotPointOnElement();
	TVector3 newHP;
	
	real32 xfix = myCommon.ivy.vertices[myCommon.ivy.FirstRootVertex].pos.x;
	if(xfix < 0) xfix += 3.0f;
	else xfix -= 3.0f;

	real32 yfix = myCommon.ivy.vertices[myCommon.ivy.FirstRootVertex].pos.z;
	if(yfix < 0) yfix += 3.0f;
	else yfix -= 3.0f;

	newHP.X() = xfix;
	newHP.Y() = yfix;
	newHP.Z() = myCommon.ivy.vertices[myCommon.ivy.FirstRootVertex].pos.y;

	instanceTree->SetHotPoint(newHP);

	TTreeTransform gmat;
	growOnMe->GetGlobalTreeTransform(gmat);

	// Insert in the scene
	growOnMe->InsertRight(instanceTree);
	
	TTreeTransform mat;
	instanceTree->GetGlobalTreeTransform(mat);
	mat.SetOffset(gmat.GetOffset());
	if(fData.usePoserScale) mat.SetUniformScaling(.09f);
	instanceTree->SetGlobalTreeTransform(mat);

	// Select the new tree element
	TMCCountedPtr<ISelectableObject> newSelectable;
	instanceTree->QueryInterface(IID_ISelectableObject, (void**)&newSelectable);
	newSelectable->AddToSelection(fSelection,true);
}




boolean	GenerateIvyDlg::Receive(int32 message, IMFResponder* source, void* data)
{
	switch (message)
	{
		case EMFPartMessage::kMsg_PartValueChanged:
		{
			if (source)
			{
				TMCCountedPtr<IMFPart> sourcePart;
				source->QueryInterface(IID_IMFPart,(void**)&sourcePart);
				ThrowIfNil(sourcePart);

				IDType partID = sourcePart->GetIMFPartID();
				TMCCountedPtr<IMFPart> parent;
				sourcePart->GetPartBaseWindow(&parent);
				TMCCountedPtr<IMFPart> fetcher;

				switch (partID)
				{
					case 'LDst': // load settings
					{
						TMCString255 DefaultFile("Load ivy settings");
						IDType Popup = '****', FileFormat = 'text';
						TMCCountedPtr<IMCFile> File;
						gFileUtilities->CreateIMCFile(&File);
						TMCArray<IDType> Types;
						Types.AddElem('text');

						if(gFileFormatUtilities->OpenFileDialog(DefaultFile, File, Types, Popup, FileFormat) == 0)
						{
							real32 rTemp;
							boolean blTemp;
							TMCfstream istream(*File, TMCiostream::in, kAnsiTranscoder);

							fetcher = parent->FindChildPartByID('ivsz');
							istream >> rTemp;
							if(fetcher) fetcher->SetValue((void *)(&rTemp), 're32', true, true);
							fetcher = parent->FindChildPartByID('ilsz');
							istream >> rTemp;
							if(fetcher) fetcher->SetValue((void *)(&rTemp), 're32', true, true);
							fetcher = parent->FindChildPartByID('ibsz');
							istream >> rTemp;
							if(fetcher) fetcher->SetValue((void *)(&rTemp), 're32', true, true);
							fetcher = parent->FindChildPartByID('mfln');
							istream >> rTemp;
							if(fetcher) fetcher->SetValue((void *)(&rTemp), 're32', true, true);
							fetcher = parent->FindChildPartByID('madd');
							istream >> rTemp;
							if(fetcher) fetcher->SetValue((void *)(&rTemp), 're32', true, true);
							fetcher = parent->FindChildPartByID('priw');
							istream >> rTemp;
							if(fetcher) fetcher->SetValue((void *)(&rTemp), 're32', true, true);
							fetcher = parent->FindChildPartByID('rndw');
							istream >> rTemp;
							if(fetcher) fetcher->SetValue((void *)(&rTemp), 're32', true, true);
							fetcher = parent->FindChildPartByID('grvw');
							istream >> rTemp;
							if(fetcher) fetcher->SetValue((void *)(&rTemp), 're32', true, true);
							fetcher = parent->FindChildPartByID('adhw');
							istream >> rTemp;
							if(fetcher) fetcher->SetValue((void *)(&rTemp), 're32', true, true);
							fetcher = parent->FindChildPartByID('brpb');
							istream >> rTemp;
							if(fetcher) fetcher->SetValue((void *)(&rTemp), 're32', true, true);
							fetcher = parent->FindChildPartByID('lfpb');
							istream >> rTemp;
							if(fetcher) fetcher->SetValue((void *)(&rTemp), 're32', true, true);
							fetcher = parent->FindChildPartByID('grst');
							istream >> rTemp;
							if(fetcher) fetcher->SetValue((void *)(&rTemp), 're32', true, true);
							fetcher = parent->FindChildPartByID('lden');
							istream >> rTemp;
							if(fetcher) fetcher->SetValue((void *)(&rTemp), 're32', true, true);
							fetcher = parent->FindChildPartByID('upos');
							istream >> blTemp;
							if(fetcher) fetcher->SetValue((void *)(&blTemp), 'bool', true, true);
						}
					}
					break;
					case 'SVst': // save settings
					{
						TMCString255 DefaultFile("Save ivy settings");
						IDType FileFormat = 'text';
						TMCCountedPtr<IMCFile> File;
						gFileUtilities->CreateIMCFile(&File);
						TMCArray<IDType> Types;
						Types.AddElem('text');

						if(gFileFormatUtilities->SaveFileDialog(DefaultFile, File, Types, FileFormat) == 0)
						{
							File->CreateOnDisk();
							TMCfstream ostream(*File, TMCiostream::out, kAnsiTranscoder);
							real32 rTemp;
							boolean blTemp;

							fetcher = parent->FindChildPartByID('ivsz');
							if(fetcher) fetcher->GetValue((void *)(&rTemp), 're32');
							ostream << rTemp << endl;
							fetcher = parent->FindChildPartByID('ilsz');
							if(fetcher) fetcher->GetValue((void *)(&rTemp), 're32');
							ostream << rTemp << endl;
							fetcher = parent->FindChildPartByID('ibsz');
							if(fetcher) fetcher->GetValue((void *)(&rTemp), 're32');
							ostream << rTemp << endl;
							fetcher = parent->FindChildPartByID('mfln');
							if(fetcher) fetcher->GetValue((void *)(&rTemp), 're32');
							ostream << rTemp << endl;
							fetcher = parent->FindChildPartByID('madd');
							if(fetcher) fetcher->GetValue((void *)(&rTemp), 're32');
							ostream << rTemp << endl;
							fetcher = parent->FindChildPartByID('priw');
							if(fetcher) fetcher->GetValue((void *)(&rTemp), 're32');
							ostream << rTemp << endl;
							fetcher = parent->FindChildPartByID('rndw');
							if(fetcher) fetcher->GetValue((void *)(&rTemp), 're32');
							ostream << rTemp << endl;
							fetcher = parent->FindChildPartByID('grvw');
							if(fetcher) fetcher->GetValue((void *)(&rTemp), 're32');
							ostream << rTemp << endl;
							fetcher = parent->FindChildPartByID('adhw');
							if(fetcher) fetcher->GetValue((void *)(&rTemp), 're32');
							ostream << rTemp << endl;
							fetcher = parent->FindChildPartByID('brpb');
							if(fetcher) fetcher->GetValue((void *)(&rTemp), 're32');
							ostream << rTemp << endl;
							fetcher = parent->FindChildPartByID('lfpb');
							if(fetcher) fetcher->GetValue((void *)(&rTemp), 're32');
							ostream << rTemp << endl;
							fetcher = parent->FindChildPartByID('grst');
							if(fetcher) fetcher->GetValue((void *)(&rTemp), 're32');
							ostream << rTemp << endl;
							fetcher = parent->FindChildPartByID('lden');
							if(fetcher) fetcher->GetValue((void *)(&rTemp), 're32');
							ostream << rTemp << endl;
							fetcher = parent->FindChildPartByID('upos');
							if(fetcher) fetcher->GetValue((void *)(&blTemp), 'bool');
							ostream << blTemp << endl;
						}
					}
					break;
				}
			}
		}
	}

	return true;
}


